# pylint: disable=protected-access, invalid-name, too-many-locals
import base64
import secrets

from cryptography import x509
from cryptography.hazmat.bindings.openssl.binding import Binding
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.serialization import pkcs7
from django.utils.crypto import salted_hmac

copenssl = Binding.lib
cffi = Binding.ffi

# SMIME isn't supported by pyca/cryptography:
# https://github.com/pyca/cryptography/issues/1621
# adjusted with latest cryptography code from https://github.com/devartis/passbook/pull/60/files
def pkcs7_sign(
    certcontent, keycontent, wwdr_certificate, data, key_password=None,
):
    """Sign data with PKCS#7.

    Args:
        keycontent (bytes): Content of pem file certificate
        keycontent (bytes): Content of key file
        wwdr_certificate (bytes): Content of Intermediate cert file
        data (bytes): Data to be signed
        key_password (bytes, optional): key file passwd. Defaults to None.
    """

    cert = x509.load_pem_x509_certificate(certcontent)
    priv_key = serialization.load_pem_private_key(keycontent, password=key_password)
    wwdr_cert = x509.load_pem_x509_certificate(wwdr_certificate)

    options = [pkcs7.PKCS7Options.DetachedSignature]
    return (
        pkcs7.PKCS7SignatureBuilder()
        .set_data(data)
        .add_signer(cert, priv_key, hashes.SHA256())
        .add_certificate(wwdr_cert)
        .sign(serialization.Encoding.DER, options)
    )


def gen_random_token():
    rand1 = secrets.token_bytes(16)
    rand2 = secrets.token_bytes(7)
    rand2 = salted_hmac(rand1, rand2).digest()
    part_a = base64.urlsafe_b64encode(rand2).rstrip(b'=').decode('ascii')
    part_b = secrets.token_urlsafe(20)
    return part_b + part_a



# Generated by Django 3.2.19 on 2023-05-25 13:39

from django.db import migrations, models
import django_walletpass.storage


class Migration(migrations.Migration):

    dependencies = [
        ('django_walletpass', '0007_auto_20190613_1807'),
    ]

    operations = [
        migrations.AlterField(
            model_name='pass',
            name='data',
            field=models.FileField(storage=django_walletpass.storage.WalletPassStorage(), upload_to='uploads/passes/%Y/%m/%d/'),
        ),
    ]



from django.db.models.signals import post_save
from django.dispatch import receiver
from django_walletpass.models import Pass


@receiver(post_save, sender=Pass)
def send_push_notification(instance=None, **_kwargs):
    instance.push_notification()



import json
from calendar import timegm

from pytz.exceptions import NonExistentTimeError

import django.dispatch
from dateutil.parser import parse
from django.db.models import Max
from django.http import HttpResponse
from django.middleware.http import ConditionalGetMiddleware
from django.shortcuts import get_object_or_404
from django.utils.http import http_date
from rest_framework import status, viewsets
from rest_framework.permissions import AllowAny
from rest_framework.response import Response
from django_walletpass.models import Log, Pass, Registration
from django_walletpass.settings import dwpconfig as WALLETPASS_CONF

# legacy constant, remove when it can be assumed no timestamps of this format are out
# there anymore
FORMAT = '%Y-%m-%d %H:%M:%S'
PASS_REGISTERED = django.dispatch.Signal()
PASS_UNREGISTERED = django.dispatch.Signal()


def get_pass(pass_type_id, serial_number):
    return get_object_or_404(Pass, pass_type_identifier=pass_type_id, serial_number=serial_number)


class RegistrationsViewSet(viewsets.ViewSet):
    """
    Gets the Serial Numbers for Passes Associated with a Device
    """
    permission_classes = (AllowAny, )

    def list(self, request, device_library_id, pass_type_id):
        passes = Pass.objects.filter(
            registrations__device_library_identifier=device_library_id,
            pass_type_identifier=pass_type_id,
        )
        if passes.count() == 0:
            return Response({}, status=status.HTTP_400_BAD_REQUEST)

        if 'passesUpdatedSince' in request.GET:
            try:
                # must be able to read UTC isoformat with TZ and FORMAT datetime string
                # as well
                date = parse(request.GET['passesUpdatedSince'])
                passes = passes.filter(updated_at__gt=date)
            except NonExistentTimeError:
                date = date.replace(hour=0, minute=0)
                passes = passes.filter(updated_at__gt=date)

        if passes:
            last_updated = passes.aggregate(Max('updated_at'))['updated_at__max']
            serial_numbers = [p.serial_number for p in passes.filter(updated_at=last_updated).all()]
            response_data = {'lastUpdated': last_updated.isoformat(), 'serialNumbers': serial_numbers}
            return Response(response_data)

        return Response({}, status=status.HTTP_204_NO_CONTENT)


# TODO: Refactor.
# - Use a ModelViewSet
# - Use auth class
class RegisterPassViewSet(viewsets.ViewSet):
    """
    get: Registers a Device to Receive Push Notifications for a Pass
    destroy: Unregister a Device
    """
    permission_classes = (AllowAny, )

    def create(self, request, device_library_id, pass_type_id, serial_number):

        pass_ = get_pass(pass_type_id, serial_number)
        if request.META.get('HTTP_AUTHORIZATION') != f"ApplePass {pass_.authentication_token}":
            return Response({}, status=status.HTTP_401_UNAUTHORIZED)
        registration = Registration.objects.filter(
            device_library_identifier=device_library_id,
            pazz=pass_,
        )
        if registration:
            return Response({}, status=status.HTTP_200_OK)
        body = json.loads(request.body)
        new_registration = Registration(
            device_library_identifier=device_library_id,
            push_token=body['pushToken'],
            pazz=pass_,
        )
        new_registration.save()
        PASS_REGISTERED.send(sender=pass_)
        return Response({}, status=status.HTTP_201_CREATED)

    def destroy(self, request, device_library_id, pass_type_id, serial_number):
        pass_ = get_pass(pass_type_id, serial_number)
        if request.META.get('HTTP_AUTHORIZATION') != f"ApplePass {pass_.authentication_token}":
            return Response({}, status=status.HTTP_401_UNAUTHORIZED)
        registration = Registration.objects.filter(
            device_library_identifier=device_library_id,
            pazz=pass_,
        )
        if registration:
            return Response({}, status=status.HTTP_200_OK)
        registration.delete()
        PASS_UNREGISTERED.send(sender=pass_)
        return Response({}, status=status.HTTP_200_OK)


class LatestVersionViewSet(viewsets.ViewSet):
    """
    get: Gets the latest version of a Pass
    """
    permission_classes = (AllowAny, )

    def retrieve(self, request, pass_type_id, serial_number):
        pass_ = get_pass(pass_type_id, serial_number)

        if request.META.get('HTTP_AUTHORIZATION') != f"ApplePass {pass_.authentication_token}":
            return Response({}, status=status.HTTP_401_UNAUTHORIZED)

        if WALLETPASS_CONF['STORAGE_HTTP_REDIRECT']:
            return Response({}, status=status.HTTP_302_FOUND, headers={'Location': pass_.data.url})

        response = HttpResponse(pass_.data.read(), content_type='application/vnd.apple.pkpass')
        response['Content-Disposition'] = 'attachment; filename=pass.pkpass'

        response['Last-Modified'] = http_date(timegm(pass_.updated_at.utctimetuple()))

        def _get_response(_request):
            return response

        return ConditionalGetMiddleware(get_response=_get_response)(request)


# TODO: use ModelViewSet
class LogViewSet(viewsets.ViewSet):
    """
    Logs messages from devices
    """
    permission_classes = (AllowAny, )

    def create(self, request):
        json_body = json.loads(request.body)
        for message in json_body['logs']:
            Log(message=message).save()
        return Response({}, status=status.HTTP_200_OK)



import asyncio
import logging
from ssl import SSLError

from aioapns import APNs, NotificationRequest
from aioapns.exceptions import ConnectionClosed
from django_walletpass.models import Registration
from django_walletpass.settings import dwpconfig as WALLETPASS_CONF

logger = logging.getLogger('walletpass.services')


class PushBackend:
    def __init__(self):
        self.loop = asyncio.get_event_loop()

    async def push_notification(self, client, token):

        try:
            request = NotificationRequest(device_token=token, message={"aps": {}},)
            await client.send_notification(request)

        except SSLError as e:
            logger.error("django_walletpass SSLError: %s", e)
        except (ConnectionError, ConnectionClosed) as e:
            logger.error("django_walletpass StreamResetError. Bad cert or token? %s", e)
        # Errors should never pass silently.
        except Exception as e:  # pylint: disable=broad-except
            # Unless explicitly silenced.
            logger.error("django_walletpass uncaught error %s", e)

    def push_notification_with_token(self, token):
        client = APNs(
            key=WALLETPASS_CONF["TOKEN_AUTH_KEY_PATH"],
            key_id=WALLETPASS_CONF["TOKEN_AUTH_KEY_ID"],
            team_id=WALLETPASS_CONF["TEAM_ID"],
            topic=WALLETPASS_CONF["PASS_TYPE_ID"],
            use_sandbox=WALLETPASS_CONF["PUSH_SANDBOX"],
        )
        return self.loop.run_until_complete(self.push_notification(client, token))

    def push_notification_from_instance(self, registration_instance):
        return self.push_notification_with_token(registration_instance.push_token)

    def push_notification_from_pk(self, registration_pk):
        registration = Registration.objects.get(pk=registration_pk)
        return self.push_notification_from_instance(registration)



from django.apps import AppConfig


class DjangoWalletpassConfig(AppConfig):
    name = 'django_walletpass'

    def ready(self):
        from django_walletpass import signals as _signals  # pylint: disable=import-outside-toplevel
